/* ARCADEmini assembler (Windows)
 * version 1.0 date 2003-09-03
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <string.h>
#include <malloc.h>

#include "id_list.h"
#include "debug.h"

//this tis the implementation of the identifier list

//anchor of doubly linked identifier list
st_id_list_entry _id_list =
{
  //this data is never used, only initialized to keep compiler happy
  /*.p_name =*/ NULL,
  /*.type =*/ id_list_unused,
  /*.firstuse_column =*/ 0,
  /*.firstuse_line =*/ 0,
  /*.decl_column =*/ 0,
  /*.decl_line =*/ 0,
  /*.p_decl_node =*/ NULL,
  /*.data =*/ { /*.unused =*/ 0 },
  //list is initially empty: point back to anchor
  /*._p_prev =*/ &_id_list,
  /*._p_next =*/ &_id_list
};

#ifdef DEBUG
//debug function: convert identifier type number to identifier type string
char * id_list_dbg_type2str( et_id_list_type type )
{
  switch( type )
  {
    case id_list_unused: return "unused";
    case id_list_con: return "con";
    case id_list_reg: return "reg";
    case id_list_addr: return "addr";
    default: return "???";
  }
}
#endif

//convert identifier type to human readable identifier type string
char * id_list_type2typestr( et_id_list_type type )
{
  switch( type )
  {
    case id_list_unused: return "unused identifier";
    case id_list_con: return "constant identifier";
    case id_list_reg: return "register identifier";
    case id_list_addr: return "address identifier";
    default: return "unknown identifier";
  }
}

//convert identifier type to prefix symbol
char * id_list_type2prefix( et_id_list_type type )
{
  switch( type )
  {
    case id_list_unused: return "?";
    case id_list_con: return "";
    case id_list_reg: return "$";
    case id_list_addr: return "@";
    default: return "?";
  }
}

//search an identifier in the list
//returns pointer to list entry or NULL if not found
st_id_list_entry * id_list_lookup( char * p_name, et_id_list_type type )
{
  st_id_list_entry * p_entry;

  //search for identifier
  for( p_entry = _id_list._p_next; p_entry != &_id_list; p_entry = p_entry->_p_next )
  {
    //identifier found and type matches
    if( strcmp( p_name, p_entry->p_name ) == 0 && type == p_entry->type )
    {
      //return pointer to entry
      debug2( "id_list_lookup: name=%s type=%s (found)", p_name, id_list_dbg_type2str( type ) );
      return p_entry;
    }
  }

  //identifier not found
  debug2( "id_list_lookup: name=%s type=%s (not found)", p_name, id_list_dbg_type2str( type ) );
  return NULL;
}

//add an identifier to the list
//does not add an identifier of the same type a second time
//initialized location of first usage and of declaration to 0, 0 when entry is created newly
//returns pointer to list entry or NULL in case of error
st_id_list_entry * id_list_add( char * p_name, et_id_list_type type )
{
  st_id_list_entry * p_entry;

  //search for identifier
  for( p_entry = _id_list._p_next; p_entry != &_id_list; p_entry = p_entry->_p_next )
  {
    //identifier found and type matches
    if( strcmp( p_name, p_entry->p_name ) == 0 && type == p_entry->type )
    {
      //return pointer to entry
      debug2( "id_list_add: name=%s type=%s (already in)", p_name, id_list_dbg_type2str( type ) );
      return p_entry;
    }
  }

  //identifier not found - we have to add it

  //allocate entry
  p_entry = (st_id_list_entry *)malloc( sizeof( st_id_list_entry ) );
  if( p_entry == NULL )
  {
    debug2( "id_list_add: name=%s type=%s (out of memory)", p_name, id_list_dbg_type2str( type ) );
    return NULL;
  }

  //allocate name string
  p_entry->p_name = strdup( p_name );
  if( p_entry->p_name == NULL )
  {
    free( p_entry );
    debug2( "id_list_add: name=%s type=%s (out of memory)", p_name, id_list_dbg_type2str( type ) );
    return NULL;
  }

  //initialize entry
  p_entry->type = type; //set type of entry
  p_entry->firstuse_column = 0; //entry was not used anywhere yet
  p_entry->firstuse_line = 0;
  p_entry->decl_column = 0; //entry was not declared anywhere yet
  p_entry->decl_line = 0;
  p_entry->p_decl_node = NULL; //no node of declaration found yet

  //initialize type data
  switch( type )
  {
    case id_list_unused:
      p_entry->data.unused = 0;
      break;
    case id_list_con:
      p_entry->data.con.val_avail = 0;
      p_entry->data.con.val = 0;
      break;
    case id_list_reg:
      p_entry->data.reg.reg_no_avail = 0;
      p_entry->data.reg.reg_no = 0;
      break;
    case id_list_addr:
      p_entry->data.addr.addr_avail = 0;
      p_entry->data.addr.addr_sec = 0;
      p_entry->data.addr.addr_ofs = 0;
      break;
  }
  
  //add entry to list
  p_entry->_p_next = &_id_list;
  p_entry->_p_prev = _id_list._p_prev;
  p_entry->_p_next->_p_prev = p_entry;
  p_entry->_p_prev->_p_next = p_entry;
        
  //return pointer to entry
  debug2( "id_list_add: name=%s type=%s (added)", p_name, id_list_dbg_type2str( type ) );
  return p_entry;
}

//enumerate identifier list
//calls a function for every entry in the identifier list
void id_list_enum( void (*enum_func)( st_id_list_entry * p_entry ) )
{
  st_id_list_entry * p_entry;

  debug( "id_list_enum" );

  //loop though all entries
  for( p_entry = _id_list._p_next; p_entry != &_id_list; p_entry = p_entry->_p_next )
    //call enumeration function
    enum_func( p_entry );
}

//free identifier list entry
void id_list_free( st_id_list_entry * p_entry )
{
  //no entry to free
  if( p_entry == NULL )
  {
    debug( "id_list_free: NULL" );
    return;
  }

  debug2( "id_list_free: name=%s type=%s", p_entry->p_name, id_list_dbg_type2str( p_entry->type ) );

  //remove node from node list
  p_entry->_p_next->_p_prev = p_entry->_p_prev;
  p_entry->_p_prev->_p_next = p_entry->_p_next;

  //free name string
  free( p_entry->p_name );

  //free entry
  free( p_entry );
}

//free entire identifier list
void id_list_free_all( )
{
  debug( "id_list_free_all" );

  //free all entries
  while( _id_list._p_next != &_id_list ) //until list is empty (anchor points to itself)
    id_list_free( _id_list._p_next ); //free first entry in list
}

